package ShellCommand::dir;

use strict;

# gonna be ls but until it becomes useful will be called dir

sub execute {
  shift;
  my @files;
  my %dirs;

  push @_, "." if ($#_ < 0);

  foreach (@_) {
    if (-d $_) {
      opendir DIR, $_;
      push @{$dirs{$_}}, readdir(DIR);
      closedir DIR;

    } elsif (-f $_) {
      push @files, $_;

    } else {
      print "ls: No such file or directory: $_\n";
    }

  }

  print_lists(\@files, \%dirs);
}

sub print_lists {
  my @files = @{$_[0]};
  my %dirs = %{$_[1]};
  my @dirnames = sort keys %dirs;
  my $first = 1;

  print_file_list(@files) if $#files >= 0;

  foreach (@dirnames) {
    if (!$first || $#files >= 0) {
      print "\n";
    }
    print "$_:\n" if ($#dirnames > 0 || $#files >= 0);
    print_file_list(@{$dirs{$_}});
  }
}

sub print_file_list {
  my @files = map { s:.*/::; $_ } sort @_;
  my $length = 0;
  my $width = 80;

  foreach (@files) {
    $length = length($_) if (length($_) > $length);
  }

  $length += 2;
  my $cols = int($width / $length);
  my $colwidth = $width / $cols;
  my $format = ("%-". $colwidth. "s") x $cols. "\n";

  my $rows = 1 + int($#files / $cols);
  for (my $i = 0; $i < $rows; $i++) {
    my @l = ();
    for (my $j = 0; $j < $cols; $j++) {
      push @l, $files[$i + $j * $rows];
    }
    printf $format, @l;
  }
}

1;
