package Timtosh::Parser;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK);

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(tokenize expand parse);
@EXPORT_OK = ();

sub tokenize {
  my $line = shift;
  my $mark = shift;

  # Only split up the line into tokens. Don't do any parsing/substitution/fancy
  # crap.
  my @elements;
  my $quote = undef;
  my $escape = 0;
  my $index = 0;
  my $markindex = 0;
  my $word = "";

  my @group = (
      '\s',
      '[\>\<\|]',
      '.',
      );

  foreach (split(//, $line)) {
    if ($escape == 2) {
      $escape = 1;
      $word .= $_;
    } elsif (/\\/) {
      $escape = 2;
      $word .= $_;
    } elsif (defined $quote) {
      # Inside some insidious quote
      if (/$quote/) {
	undef $quote;
      }
      $word .= $_;
    } else {
      my ($oldgroup, $newgroup);
      if ($escape == 1) {
	$escape = 0;
	$newgroup = $oldgroup = -1;
      } else {
	for (my $i = 0; $i <= $#group; $i++) {
	  if (!defined($oldgroup) && substr($word, -1) =~ /$group[$i]/) {
	    $oldgroup = $i;
	  }
	  if (!defined($newgroup) && /$group[$i]/) {
	    $newgroup = $i;
	  }
	}
      }
      #print "old: $oldgroup (". substr($word, -1). ")\n";
      #print "new: $newgroup ($_)\n";

      # for echo "a"   (don't want to echo " a")
      if ($oldgroup != $newgroup) {
	if (/[\"\']/) {
	  $quote = $_;
	}

	if (length($word)) {
	  push @elements, $word;
	  if ($index <= $mark) {
	    $markindex++;
	  }
	}
	$word = $_;

      } elsif (/[\"\']/) {
	$quote = $_;
	$word .= $_;

      } else {
	$word .= $_;
      }
    }

    $index++;
  }

  if ($word) {
    push @elements, $word;
  }
  #print "elements: ", join(",", @elements), "\n";

  my @tokens;
  while ($_ = shift @elements) {
    if ($_ eq '|') {
      push @tokens, [ $_, 'PIPE' ];

    } elsif (/([12])>/) {
      push @tokens, [ $_, 'REDIRECT', $1, '>'];

    } elsif ($_ eq '>') {
      push @tokens, [ $_, 'REDIRECT', 1, '>'];

    } elsif ($_ eq '<') {
      push @tokens, [ $_, 'REDIRECT', 0, '<'];

    } elsif ($_ eq ';') {
      push @tokens, [ $_, 'END' ];

    } elsif ($_ eq '&') {
      push @tokens, [ $_, 'BACKGROUND' ], [ 'END' ];

    } elsif ($_ =~ /^\s*$/) {
      push @tokens, [ $_, 'WHITESPACE' ];

    } else {
      push @tokens, [ $_, 'WORD', $_ ];
    }
  }

#  if ($word) {
#    push @tokens, [ 'UNFINISHED', $word ];
#  }

  #print_tokens("Tokens", @tokens);

  if (defined($mark)) {
    return $markindex, @tokens;
  } else {
    return @tokens;
  }
}

sub expand_glob {
  my ($re, $word);

  my $quote;
  my $escape = 0;
  # $first includes a . if something like .* or {a,.}* happens
  my $first;
  foreach (split(//, $_[0])) {
    if ($quote) {
      # inside some quote
      if ($escape) {
	$escape = 0;
	if (/$quote/ || ($quote eq '"' && /\\/)) {
	  $re .= "\\$_";
	  $word .= $_;
	} else {
	  $re .= "\\$_";
	  $word .= "\\$_";
	}
      } elsif (/\\/) {
	$escape = 1;
      } elsif (/$quote/) {
	undef $quote;
      } elsif (/\w/) {
	$re .= $_;
	$word .= $_;
      } else {
	$re .= "\\$_";
	$word .= $_;
      }
    } else {
      # not in a quote
      if ($escape == 1) {
	$escape = 0;
	$re .= "\\$_";
	$word .= $_;
      } elsif (/\\/) {
	$escape = 1;
      } elsif (/[\'\"]/) {
	$quote = $_;
      } elsif (/\*/) {
	$re .= ".*";
	$word .= "*";
      } elsif (/\?/) {
	$re .= ".";
	$word .= "?";
      } elsif (/\w/) {
	$re .= $_;
	$word .= $_;
      } else {
	$re .= "\\$_";
	$word .= $_;
      }
    }

    # won't work once I implement foo{bar,baz}
    $first ||= substr($word, 0, 1);
  }

  my @matches = split(/\\\//, $re);
#  if (substr($re, -2) eq '\/') {
#    push @matches, "";
#  }

  my @files = matching_files(undef, 1, @matches);
  #print "matched: ", join(", ", @files), "\n";

  return @files if $#files >= 0;
  return $word;
}

sub matching_files {
  #print "matching_files(", join(", ", @_), ")\n";
  my $dir = shift;
  my $match_dot = shift;
  my @re = @_;

  my $re = shift @re;

  if (length($re) == 0) {
    return matching_files("/", $match_dot, @re);
  }

  my $exclude;
  if ($match_dot) {
    $exclude = '^\.\.?$';
  } else {
    $exclude = '^\.';
  }

  opendir DIR, ($dir || ".");
  my @files = readdir(DIR);
  closedir DIR;

  if ($dir && $dir ne '/') {
    $dir = "$dir/";
  }

  @files = grep(/^$re$/, @files);
  if ($#files > 0) {
    # get rid of . and ..
    @files = grep(!/^\.\.?$/, @files);
  }

  foreach (@files) {
    s/^/$dir/;
  }

  if ($#re >= 0) {
    # more levels left. Only keep dirs
    #print "Only keep directories in ", join(", ", @files), "\n";
    @files = grep { -d $_ } @files;
    #print "Only kept directories ", join(", ", @files), "\n";
  }

  #print "matched 2: ", join(", ", @files), "\n";

  if ($#re >= 0) {
    my @results;
    foreach (@files) {
      push @results, matching_files($_, $match_dot, @re);
    }
    return @results;

  } else {
    my @results;
    foreach (@files) {
      push @results, $_;
    }
    return @results;
  }
}

# expand globs, get rid of quotes, unescape stuff, ...
# get rid of all whitespace tokens
sub expand {
  my @result;
  my $a;

  #push @result, (shift shift);
  while ($a = shift) {
    shift @$a;
    next if $a->[0] eq 'WHITESPACE';
    if ($a->[0] eq 'WORD') {
      my $word = $a->[1];
      my @files = expand_glob($word);
      my @words;
      foreach (@files) {
	push @words, [ 'WORD', $_ ];
      }
      push @result, @words;
    } else {
      push @result, $a;
    }
  }

  #print_tokens("after expansion", @result);

  return @result;
}

# Line := (ComplexCommand END?)+
sub parseLine {
  my @tokens = @_;
  my @elements;
  my @element;
  my $element;

  while ($#tokens >= 0) {
    ($element, @tokens) = parseComplexCommand(@tokens);
    if (!defined $element) {
      return undef;
    }

    if ($#tokens >= 0 && $tokens[0]->[0] eq 'END') {
      shift @tokens;
    }

    push @elements, $element;
  }

  if ($#elements > 0) {
    return [ \&::sequentially, @elements ];
  }
  
  return $elements[0];
}

# ComplexCommand := SimpleCommand (PIPE ComplexCommand)* BACKGROUND?
sub parseComplexCommand {
  my @tokens = @_;
  my @simplecommands;

  #print_tokens("parseComplexCommand", @tokens);

  ($simplecommands[0], @tokens) = parseSimpleCommand(@tokens);
  #print_tokens("parseComplexCommand remaining", @tokens);

  while ($#tokens >= 0 && $tokens[0]->[0] eq 'PIPE') {
    shift @tokens;
    my $sc;
    ($sc, @tokens) = parseSimpleCommand(@tokens);
    push @simplecommands, $sc;
  }

  my $foreground = 1;
  if ($#tokens >= 0 && $tokens[0]->[0] eq 'BACKGROUND') {
    shift @tokens;
    $foreground = 0;
  }

  return [ \&::exec_complex_command, [ @simplecommands ], $foreground ],
      @tokens;
}

# SimpleCommand := VAR* WORD (WORD | (REDIRECT WORD))+
sub parseSimpleCommand {
  my @tokens = @_;

  my @words;
  my @options;

  #print_tokens("parseSimpleCommand", @tokens);

  my $token = shift @tokens;
  push @words, $token->[1];
  while ($#tokens >= 0 &&
      ($tokens[0]->[0] eq 'WORD' || $tokens[0]->[0] eq 'REDIRECT')) {
    my $token = shift @tokens;
    if ($token->[0] eq 'WORD') {
      push @words, $token->[1];
    } elsif ($token->[0] eq 'REDIRECT') {
      my $file = shift @tokens;
      die "no word: ". join(", ", @$file) unless $file->[0] eq 'WORD';
      push @options, [ @$token, $file->[1] ];
    } else {
      die "WTF!";
    }
  }

  #print_tokens("parseSimpleCommand returning $#tokens", @tokens);

  return [ \&::command, \@options, @words ], @tokens;
}

sub parse {
  return parseLine(@_);
}

sub print_tokens {
  print "$_[0]: ";
  shift;
  foreach (@_) {
    print "(", join(", ", @{$_}), ")";
  }
  print "\n";
}

1;
