package Timtosh::Prompt;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK);

require Exporter;
use Term::ReadKey;
use Timtosh::Complete;

@ISA = qw(Exporter);
@EXPORT = qw(prompt);
@EXPORT_OK = ();

sub prompt_string {
  $_ = shift || 'timtosh%# ';

  # %n -> username
  s/%n/$Timtosh::shelldata{'username'}/g;

  # %# -> # if we're root, % otherwise
  if ($> == 0) {
    s/%#/#/g;
  } else {
    s/%#/%/g;
  }

  # %m -> machine name
  s/%m/$Timtosh::shelldata{'machinename'}/g;

  # %~ -> current directory, with ~ substituted for user's homedir, and
  # ~mofo substituted for mofo's home dir (TODO)
  my $dir = $ENV{'PWD'};
  $dir =~ s/^$Timtosh::shelldata{'homedir'}/~/;
  s/%~/$dir/;

  s/%\$/$$/g;

  return $_;
}

# Takes a string and shows all non-printable stuff so it's still readable.
sub legibly {
  $_ = $_[0];

  # Properly show ^@, ^A - ^Z, ^[, ^\, ^]
  s/([\000-\037])/"^".chr(ord($1) + 64)/ge;

  return $_;
}

sub prompt {
  my $string = "";
  my $c;
  my $loc = 0;	# location in command string
  my $pos;	# position on screen

  my $term = shift;
  my $prompt = prompt_string(shift);

  ReadMode 4;

  print $prompt;
  my $pos = 0;

  while (($c = ReadKey(0)) ne "\n") {
    if (ord($c) >= ord(" ") && ord($c) <= ord("~")) {
      # something non-printable
      $string = substr($string, 0, $loc). $c. substr($string, $loc);
      print $term->Tputs('im'), legibly($c), $term->Tputs('ei');
      $loc++;

    } elsif ($c eq "\b") {	# Backspace (^H)
      print "". (($term->Tputs('bc'). $term->Tputs('dc')) x
	  length(legibly(chop $string)));
      $loc--;

    } elsif (ord($c) == 0) {	# NULL
      # Just get out.
      return undef;

    } elsif (ord($c) == 1) {	# ^A
      # Go to the front of the line.
      print $term->Tputs('bc') x length(legibly(substr($string, 0, $loc)));
      $loc = 0;

    } elsif (ord($c) == 2) {	# ^B
      # Go back one character.
      if ($loc > 0) {
	$loc--;
	print $term->Tputs('bc') x length(legibly(substr($string, $loc, 1)));
      } else {
	Timtosh::beep();
      }

    } elsif (ord($c) == 3) {	# ^C
      print "\n";
      return "";

    } elsif (ord($c) == 4) {	# ^D
      # EOF on a blank line.
      return undef if (length($string) == 0);
      # delete a character
      if ($loc < length($string)) {
	$string = substr($string, 0, $loc). substr($string, $loc + 1);
	print $term->Tputs('dc');
      }

    } elsif (ord($c) == 5) {	# ^E
      print $term->Tputs('nd') x (length($string) - $loc);
      $loc = length($string);

    } elsif (ord($c) == 6) {	# ^F
      if ($loc < length($string)) {
	$loc++;
	print $term->Tputs('nd');
      }

    } elsif (ord($c) == 9) {	# Tab
      print $term->Tputs('bc') x length(legibly(substr($string, 0, $loc)));
      ($string, $loc) = complete($string, $loc);
      #print $term->Tputs('ce'), $string;
      print "\r$prompt". legibly($string);
      my $diff = length(legibly($string)) - $loc;
      print $term->Tputs('le') x $diff;

    #} elsif (ord($c) == 21) {	# ^U
    #  print $term->Tputs('bc') x length($string), $term->Tputs('ce');
    #  $string = "";

    } elsif (ord($c) == 22) {	# ^V
      my $char = ReadKey(0);
      $string .= $char;
      print legibly($char);
      $loc++;

    } else {

      print "\nDon't know what to do with character ", ord($c), "\n$string";
    }
  }
  print "\n";

  ReadMode 0;

  return "$string$c";
}

1;
